from .__about__ import (
    __license__,
    __copyright__,
    __url__,
    __contributors__,
    __version__,
    __doc__
)

__all__ = [
    "__license__",
    "__copyright__",
    "__url__",
    "__contributors__",
    "__version__",
    "__doc__"
]
